#!/usr/bin/env python
# Created by "Thieu" at 10:49, 01/07/2022 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import numpy as np


def rounder(x, condition):
    ndim = len(x)
    temp_2x = 2 * x
    for idx in range(0, ndim):
        dec, inter = np.modf(temp_2x[idx])
        if temp_2x[idx] <= 0 and dec >= 0.5:
            temp_2x[idx] = inter - 1
        elif dec < 0.5:
            temp_2x[idx] = inter
        else:
            temp_2x[idx] = inter + 1
    temp_2x = temp_2x / 2
    return np.where(condition < 0.5, x, temp_2x)


def griewank_func(x):
    x = np.array(x).ravel()
    t1 = np.sum(x**2) / 4000
    t2 = np.prod([np.cos(x[idx] / np.sqrt(idx+1)) for idx in range(0, len(x))])
    return t1 - t2 + 1


def rosenbrock_func(x):
    x = np.array(x).ravel()
    return np.sum([100*(x[idx]**2 - x[idx+1])**2 + (x[idx] - 1)**2 for idx in range(0, len(x)-1)])


def scaffer_func(x):
    x = np.array(x).ravel()
    return 0.5 + (np.sin(np.sqrt(np.sum(x**2)))**2 - 0.5) / (1 + 0.001 * np.sum(x**2))**2


def rastrigin_func(x):
    x = np.array(x).ravel()
    return np.sum(x**2 - 10*np.cos(2*np.pi*x) + 10)


def weierstrass_func(x, a=0.5, b=3., k_max=20):
    x = np.array(x).ravel()
    ndim = len(x)
    k = np.arange(0, k_max+1)
    result = 0
    for idx in range(0, ndim):
        result += np.sum(a**k * np.cos(2*np.pi*b**k*(x[idx] + 0.5)))
    return result - ndim * np.sum(a**k * np.cos(np.pi*b**k))


def ackley_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    t1 = np.sum(x**2)
    t2 = np.sum(np.cos(2*np.pi*x))
    return -20*np.exp(-0.2 * np.sqrt(t1 / ndim)) - np.exp(t2 / ndim) + 20 + np.e


def sphere_func(x):
    x = np.array(x).ravel()
    return np.sum(x**2)


def rotated_expanded_scaffer_func(x):
    x = np.array(x).ravel()
    results = [scaffer_func([x[idx], x[idx+1]]) for idx in range(0, len(x)-1)]
    return np.sum(results) + scaffer_func([x[-1], x[0]])


def f8f2_func(x):
    x = np.array(x).ravel()
    results = [griewank_func(rosenbrock_func([x[idx], x[idx+1]])) for idx in range(0, len(x) - 1)]
    return np.sum(results) + griewank_func(rosenbrock_func([x[-1], x[0]]))


def non_continuous_expanded_scaffer_func(x):
    x = np.array(x).ravel()
    y = rounder(x, np.abs(x))
    results = [scaffer_func([y[idx], y[idx+1]]) for idx in range(0, len(x) - 1)]
    return np.sum(results) + scaffer_func([y[-1], y[0]])


def non_continuous_rastrigin_func(x):
    x = np.array(x).ravel()
    y = rounder(x, np.abs(x))
    results = [rastrigin_func([y[idx], y[idx + 1]]) for idx in range(0, len(x) - 1)]
    return np.sum(results) + rastrigin_func([y[-1], y[0]])


def elliptic_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    idx = np.arange(0, ndim)
    return np.sum((10**6)**(idx/(ndim-1)) * x**2)


def sphere_noise_func(x):
    x = np.array(x).ravel()
    return np.sum(x**2)*(1 + 0.1 * np.abs(np.random.normal(0, 1)))


def twist_func(x):
    # This function in CEC-2008 F7
    return 4 * (x ** 4 - 2 * x ** 3 + x ** 2)


def doubledip(x, c, s):
    # This function in CEC-2008 F7
    if -0.5 < x < 0.5:
        return (-6144 * (x - c) ** 6 + 3088 * (x - c) ** 4 - 392 * (x - c) ** 2 + 1) * s
    else:
        return 0


def fractal_1d_func(x):
    # This function in CEC-2008 F7
    np.random.seed(0)
    result1 = 0.0
    for k in range(1, 4):
        result2 = 0.0
        upper = 2 ** (k-1) + 1
        for t in range(1, upper):
            selected = np.random.choice([0, 1, 2], p=1/3*np.ones(3))
            result2 += np.sum([doubledip(x, np.random.uniform(0, 1), 1.0/(2**(k-1) * (2 - np.random.uniform(0, 1)))) for _ in range(0, selected)])
        result1 += result2
    return result1


def schwefel_12_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    return np.sum([np.sum(x[:idx])**2 for idx in range(0, ndim)])


def tosz_func(x):
    def transform(xi):
        if xi > 0:
            c1, c2, x_sign = 10., 7.9, 1.0
            x_star = np.log(np.abs(xi))
        elif xi == 0:
            c1, c2, x_sign, x_star = 5.5, 3.1, 0., 0.
        else:
            c1, c2, x_sign = 5.5, 3.1, -1.
            x_star = np.log(np.abs(xi))
        return x_sign * np.exp(x_star + 0.049 * (np.sin(c1*x_star) + np.sin(c2*x_star)))

    x = np.array(x).ravel()
    x[0] = transform(x[0])
    x[-1] = transform(x[-1])
    return x


def tasy_func(x, beta=0.5):
    x = np.array(x).ravel()
    ndim = len(x)
    idx = np.arange(0, ndim)
    up = 1 + beta * ((idx - 1) / (ndim-1)) * np.sqrt(np.abs(x))
    x_temp = np.abs(x) ** up
    return np.where(x > 0, x_temp, x)


def bent_cigar_func(x):
    x = np.array(x).ravel()
    return x[0]**2 + 10**6 * np.sum(x[1:]**2)


def discus_func(x):
    x = np.array(x).ravel()
    return 10**6 * x[0]**2 + np.sum(x[1:]**2)


def different_powers_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    idx = np.arange(0, ndim)
    up = 2 + 4*idx/(ndim-1)
    return np.sqrt(np.sum(np.abs(x)**up))


def generate_diagonal_matrix(size, alpha=10):
    idx = np.arange(0, size)
    diagonal = alpha**(idx / (2*(size-1)))
    matrix = np.zeros((size, size), float)
    np.fill_diagonal(matrix, diagonal)
    return matrix


def gz_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    t1 = (500 - np.mod(x, 500)) * np.sin(np.sqrt(np.abs(500 - np.mod(x, 500)))) - (x - 500)**2/(10000*ndim)
    t2 = (np.mod(np.abs(x), 500) - 500) * np.sin(np.sqrt(np.abs(np.mod(np.abs(x), 500) - 500))) - (x+500)**2/(10000*ndim)
    t3 = x*np.sin(np.abs(x)**0.5)
    conditions = [x < -500, (-500 <= x) & (x <= 500), x > 500]
    choices = [t2, t3, t1]
    y = np.select(conditions, choices, default=np.nan)
    # y = x.copy()
    # for idx in range(0, ndim):
    #     if x[idx] > 500:
    #         y[idx] = (500 - np.mod(x[idx], 500)) * np.sin(np.sqrt(np.abs(500 - np.mod(x[idx], 500)))) - (x[idx] - 500)**2/(10000*ndim)
    #     elif x[idx] < -500:
    #         y[idx] = (np.mod(x[idx], 500) - 500) * np.sin(np.sqrt(np.abs(np.mod(np.abs(x[idx]), 500) - 500))) - (x[idx]+500)**2/(10000*ndim)
    #     else:
    #         y[idx] = x[idx]*np.sin(np.abs(x[idx])**0.5)
    return y


def katsuura_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    result = 1.0
    for idx in range(0, ndim):
        temp = np.sum([ np.abs(2**j * x[idx] - np.round(2**j * x[idx])) / 2**j for j in range(1, 33)])
        result *= (1 + (idx+1)*temp) ** (10.0 / ndim**1.2)
    return (result - 1)*10 / ndim**2


def lunacek_bi_rastrigin_func(x, z, miu0=2.5, d=1.):
    x = np.array(x).ravel()
    ndim = len(x)
    s = 1 - 1.0 / (2 * np.sqrt(ndim + 20) - 8.2)
    miu1 = -np.sqrt((miu0**2 - d)/s)
    temp1 = np.sum((x - miu0)**2)
    temp2 = d*ndim + s*np.sum((x - miu1)**2)
    result1 = min(temp1, temp2)
    return result1 + 10*(ndim - np.sum(np.cos(2*np.pi*z)))


def calculate_weight(x, xichma=1.):
    ndim = len(x)
    weight = 1
    temp = np.sum(x ** 2)
    if temp != 0:
        weight = (1.0 / np.sqrt(temp)) * np.exp(-temp / (2 * ndim * xichma ** 2))
    return weight


def modified_schwefel_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    z = x + 4.209687462275036e+002
    return 418.9829 * ndim - np.sum(gz_func(z))


def happy_cat_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    t1 = np.sum(x)
    t2 = np.sum(x**2)
    return np.abs(t2 - ndim)**0.25 + (0.5*t2 + t1) / ndim + 0.5


def hgbat_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    t1 = np.sum(x)
    t2 = np.sum(x**2)
    return np.abs(t2**2 - t1**2)**0.5 + (0.5*t2 + t1) / ndim + 0.5


def zakharov_func(x):
    x = np.array(x).ravel()
    temp = np.sum(0.5*x)
    return np.sum(x**2) + temp**2 + temp**4


def levy_func(x):
    x = np.array(x).ravel()
    w = 1 + (x - 1) / 4
    t1 = np.sin(np.pi*w[0])**2 + (w[-1] - 1)**2 * (1 + np.sin(2*np.pi*w[-1])**2)
    t2 = np.sum((w[:-1] - 1)**2 * (1 + 10*np.sin(np.pi*w[:-1] + 1)**2))
    return t1 + t2


def schaffer_f7_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    result = 0.0
    for idx in range(0, ndim-1):
        t = x[idx]**2 + x[idx+1]**2
        result += np.sqrt(t) * (np.sin(50.*t**0.2) + 1)
    return (result/(ndim-1))**2


def storn_chebyshev_polynomial_fitting_func(x, d=72.661):
    x = np.array(x).ravel()
    ndim = len(x)
    m = 32*ndim
    j1 = np.arange(0, ndim)
    upper = ndim - j1

    u = np.sum(x * 1.2**upper)
    v = np.sum(x * (-1.2) ** upper)
    p1 = 0 if u >= d else (u-d)**2
    p2 = 0 if v >=d else (v-d)**2

    wk = np.array([np.sum(x * (2.*k/m - 1)**upper) for k in range(0, m+1)])
    conditions = [wk < 1, (1 <= wk) & (wk <= 1), wk > 1]
    t1 = (wk + 1)**2
    t2 = np.zeros(len(wk))
    t3 = (wk - 1)**2
    choices = [t1, t2, t3]
    pk = np.select(conditions, choices, default=np.nan)
    p3 = np.sum(pk)
    return p1 + p2 +p3


def inverse_hilbert_matrix_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    n = int(np.sqrt(ndim))
    I = np.identity(n)
    H = np.zeros((n, n))
    Z = np.zeros((n, n))
    for i in range(0, n):
        for k in range(0, n):
            Z[i, k] = x[i + n*k]
            H[i, k] = 1./(i+k+1)
    W = np.matmul(H,Z) - I
    return np.sum(W)


def lennard_jones_minimum_energy_cluster_func(x):
    x = np.array(x).ravel()
    ndim = len(x)
    result = 12.7120622568
    n_upper = int(ndim/3)
    for i in range(1, n_upper):
        for j in range(i+1, n_upper+1):
            idx1, idx2 = 3*i, 3*j
            dij = ((x[idx1-3] - x[idx2-3])**2 + (x[idx1-2] - x[idx2-2])**2 + (x[idx1-1] - x[idx2-1])**2)**3
            if dij == 0:
                result += 0
            else:
                result += (1./dij**2 - 2./dij)
    return result


expanded_griewank_rosenbrock_func = f8f2_func
expanded_scaffer_f6_func = rotated_expanded_scaffer_func



#!/usr/bin/env python
# Created by "Thieu" at 16:39, 05/05/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import numpy as np
from abc import ABC, abstractmethod


class Engineer(ABC):
    """
    Defines an abstract class for engineering design problems.

    All subclasses should implement the ``evaluate`` method for a particular optimization problem.

    Attributes
    ----------
    bounds : list
        The lower/upper bounds of the problem. This a 2D-matrix of [lower, upper] array that contain the lower and upper bounds.
        By default, each problem has its own bounds. But user can try to put different bounds to test the problem.
    n_dims : int
        The dimensionality of the problem. It is calculated from bounds
    lb : np.ndarray
        The lower bounds for the problem
    ub : np.ndarray
        The upper bounds for the problem
    f_global : float
        The global optimum of the evaluated function.
    x_global : np.ndarray
        A list of vectors that provide the locations of the global minimum.
        Note that some problems have multiple global minima, not all of which may be listed.
    n_fe : int
        The number of function evaluations that the object has been asked to calculate.
    dim_changeable : bool
        Whether we can change the benchmark function `x` variable length (i.e., the dimensionality of the problem)
    """

    name = "Benchmark name"
    latex_formula = r'f(\mathbf{x})'
    latex_formula_dimension = r'd \in \mathbb{N}_{+}^{*}'
    latex_formula_bounds = r'x_i \in [-2\pi, 2\pi], \forall i \in \llbracket 1, d\rrbracket'
    latex_formula_global_optimum = r'f(0, ..., 0)=-1, \text{ for}, m=5, \beta=15'
    continuous = True
    linear = False
    convex = True
    unimodal = False
    separable = False

    differentiable = True
    scalable = True
    randomized_term = False
    parametric = True

    modality = True  # Number of ambiguous peaks, unknown # peaks
    # n_basins = 1
    # n_valleys = 1

    def __init__(self):
        self._bounds = None
        self._n_dims = None
        self._n_objs = 1
        self._n_cons = 0
        self._n_ineq_cons = -1
        self._n_eq_cons = -1

        self.f_penalty = None
        self.f_global = None
        self.x_global = None
        self.n_fe = 0
        self.paras = {}
        self.epsilon = 1e-8
        self.w = 1e8

    def get_objs(self, x):
        """
        Compute the values of the objective functions for a given set of input values.
        """
        pass

    def get_cons(self, x):
        """
        Compute the values of the constraint functions for a given set of input values.
        """
        pass

    def get_eq_cons(self, x):
        """
        Compute the values of the equality constraint functions for a given set of input values.
        """
        pass

    def get_ineq_cons(self, x):
        """
        Compute the values of the inequality constraint functions for a given set of input values.
        """
        pass

    def get_paras(self):
        """
        Return the parameters of the problem. Depended on function
        """
        default = {"bounds": self._bounds, "n_dims": self._n_dims, }
        return {**default, **self.paras}

    @property
    def bounds(self):
        """
        The lower/upper bounds to be used for optimization problem. This a 2D-matrix of [lower, upper] array that contain the lower and upper
        bounds for the problem. The problem should not be asked for evaluation outside these bounds. ``len(bounds) == n_dims``.
        """
        return self._bounds

    @property
    def n_dims(self):
        """
        The dimensionality of the problem.
        """
        return self._n_dims

    @property
    def n_objs(self):
        """
        The number of objective functions of the problem.
        """
        return self._n_objs

    @property
    def n_cons(self):
        """
        The number of constraint functions of the problem.
        """
        return self._n_cons

    @property
    def n_eq_cons(self):
        """
        The number of equality constraint functions of the problem.
        """
        return self._n_eq_cons

    @property
    def n_ineq_cons(self):
        """
        The number of inequality constraint functions of the problem.
        """
        return self._n_ineq_cons

    @property
    def lb(self):
        """
        The lower bounds for the problem

        Returns
        -------
        lb : 1D-vector
            The lower bounds for the problem
        """
        return np.array([x[0] for x in self.bounds])

    @property
    def ub(self):
        """
        The upper bounds for the problem

        Returns
        -------
        ub : 1D-vector
            The upper bounds for the problem
        """
        return np.array([x[1] for x in self.bounds])

    def amend_position(self, x, lb=None, ub=None):
        """
        Amend position to fit the format of the problem

        Parameters
        ----------
        x : np.ndarray
            The current position (solution)
        """
        return x

    def create_solution(self):
        """
        Create a random solution for the current problem

        Returns
        -------
        solution: np.ndarray
            The random solution
        """
        return np.random.uniform(self.lb, self.ub)

    def check_solution(self, x):
        """
        Raise the error if the problem size is not equal to the solution length

        Parameters
        ----------
        x : np.ndarray
            The solution
        """
        if len(x) != self._n_dims:
            raise ValueError(f"The length of solution should has {self._n_dims} variables!")

    def default_penalty(self, list_objs=None, list_cons=None):
        list_objs_new = np.zeros_like(list_objs)
        for idx, val in enumerate(list_objs):
            temp = val + self.w * np.sum([max(0, f_con) for f_con in list_cons])
            list_objs_new[idx] = temp
        return list_objs_new

    def check_penalty_func(self, func=None):
        if callable(func):
            self.f_penalty = func
        else:
            self.f_penalty = self.default_penalty

    def evaluate(self, x):
        """
        Evaluation of the benchmark function.

        Parameters
        ----------
        x : np.ndarray, list, tuple
            The candidate vector for evaluating the benchmark problem. Must have ``len(x) == self.n_dims``.

        Returns
        -------
        val : float
              the evaluated benchmark function
        """
        pass



#!/usr/bin/env python
# Created by "Thieu" at 13:58, 09/05/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import numpy as np


class LabelEncoder:
    """
    Encode categorical features as integer labels.
    """

    def __init__(self):
        self.unique_labels = None
        self.label_to_index = {}

    def fit(self, y):
        """
        Fit label encoder to a given set of labels.

        Parameters:
        -----------
        y : array-like
            Labels to encode.
        """
        self.unique_labels = np.unique(y)
        self.label_to_index = {label: i for i, label in enumerate(self.unique_labels)}

    def transform(self, y):
        """
        Transform labels to encoded integer labels.

        Parameters:
        -----------
        y : array-like
            Labels to encode.

        Returns:
        --------
        encoded_labels : array-like
            Encoded integer labels.
        """
        if self.unique_labels is None:
            raise ValueError("Label encoder has not been fit yet.")
        return np.array([self.label_to_index[label] for label in y])

    def fit_transform(self, y):
        """Fit label encoder and return encoded labels.

        Parameters
        ----------
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        y : array-like of shape (n_samples,)
            Encoded labels.
        """
        self.fit(y)
        return self.transform(y)

    def inverse_transform(self, y):
        """
        Transform integer labels to original labels.

        Parameters:
        -----------
        y : array-like
            Encoded integer labels.

        Returns:
        --------
        original_labels : array-like
            Original labels.
        """
        if self.unique_labels is None:
            raise ValueError("Label encoder has not been fit yet.")
        return np.array([self.unique_labels[i] if i in self.label_to_index.values() else "unknown" for i in y])



#!/usr/bin/env python
# Created by "Thieu" at 13:24, 27/02/2022 ----------%
#       Email: nguyenthieu2102@gmail.com            %
#       Github: https://github.com/thieu1995        %
# --------------------------------------------------%

from setuptools import setup, find_packages


def readme():
    with open('README.md', encoding='utf-8') as f:
        README = f.read()
    return README


setup(
    name="enoppy",
    version="0.1.0",
    author="Thieu",
    author_email="nguyenthieu2102@gmail.com",
    description="ENOPPY: A Python Library for Engineering Optimization Problems",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords=["optimization", "metaheuristics", "MHA", "mathematical optimization", "nature-inspired algorithms",
              "evolutionary computation", "soft computing", "population-based algorithms",
              "Stochastic optimization", "Global optimization", "Convergence analysis", "Search space exploration",
              "Local search", "Computational intelligence", "Black-box optimization", "Robust optimization",
              "Hybrid algorithms", "Benchmark functions", "Metaheuristic design", "Performance analysis",
              "Exploration versus exploitation", "Self-adaptation", "Constrained optimization",
              "Intelligent optimization", "Adaptive search", "Simulations", "Algorithm selection"],
    url="https://github.com/thieu1995/enoppy",
    project_urls={
        'Documentation': 'https://enoppy.readthedocs.io/',
        'Source Code': 'https://github.com/thieu1995/enoppy',
        'Bug Tracker': 'https://github.com/thieu1995/enoppy/issues',
        'Change Log': 'https://github.com/thieu1995/enoppy/blob/master/ChangeLog.md',
        'Forum': 'https://t.me/+fRVCJGuGJg1mNDg1',
    },
    packages=find_packages(exclude=['tests*', 'examples*']),
    include_package_data=True,
    license="GPLv3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: System :: Benchmark",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    install_requires=["numpy>=1.16.5", "matplotlib>=3.3.0", "scipy>=1.7.1", "mealpy>=2.5.3", "opfunu>=1.0.0"],
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.1"],
    },
    python_requires='>=3.7',
)



#!/usr/bin/env python
# Created by "Thieu" at 11:23, 16/03/2020 ----------%
#       Email: nguyenthieu2102@gmail.com            %
#       Github: https://github.com/thieu1995        %
# --------------------------------------------------%
#
# Examples:
# >>> from enoppy.paper_based import moeosma_2023
#
# >>> p1 = moeosma_2023.SpeedReducerProblem()
#
# >>> print(p1.get_paras())
# >>> print(p1.bounds)
# >>> print(p1.n_dims)
# >>> print(p1.n_objs)
# >>> print(p1.n_cons)
# >>> print(p1.lb)
# >>> print(p1.ub)
#
# >>> x0 = p1.create_solution()
# >>> print(x0)
# >>> x0 = p1.amend_position(x0)
#
# >>> print(p1.get_objs(x0)) # Get all objectives
# >>> print(p1.get_cons(x0)) # Get all constraints
# >>> print(p1.evaluate(x0)) # Evaluate the x0 (calculate fitness value with objectives and constraints combination)

__version__ = "0.1.0"

import inspect
import re
from .utils import *



#!/usr/bin/env python
# Created by "Thieu" at 21:39, 29/06/2022 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

import operator
import numpy as np


def is_in_bound(value, bound):
    ops = None
    if type(bound) is tuple:
        ops = operator.lt
    elif type(bound) is list:
        ops = operator.le
    if bound[0] == float("-inf") and bound[1] == float("inf"):
        return True
    elif bound[0] == float("-inf") and ops(value, bound[1]):
        return True
    elif ops(bound[0], value) and bound[1] == float("inf"):
        return True
    elif ops(bound[0], value) and ops(value, bound[1]):
        return True
    return False


def is_str_in_list(value: str, my_list: list):
    if type(value) == str and my_list is not None:
        return True if value in my_list else False
    return False


def check_int(name: str, value: int, bound=None):
    if type(value) in [int, float]:
        if bound is None:
            return int(value)
        elif is_in_bound(value, bound):
            return int(value)
    bound = "" if bound is None else f"and value should be in range: {bound}"
    raise ValueError(f"'{name}' is an integer {bound}.")


def check_float(name: str, value: int, bound=None):
    if type(value) in [int, float]:
        if bound is None:
            return float(value)
        elif is_in_bound(value, bound):
            return float(value)
    bound = "" if bound is None else f"and value should be in range: {bound}"
    raise ValueError(f"'{name}' is a float {bound}.")


def check_str(name: str, value: str, bound=None):
    if type(value) is str:
        if bound is None or is_str_in_list(value, bound):
            return value
    bound = "" if bound is None else f"and value should be one of this: {bound}"
    raise ValueError(f"'{name}' is a string {bound}.")


def check_bool(name: str, value: bool, bound=(True, False)):
    if type(value) is bool:
        if value in bound:
            return value
    bound = "" if bound is None else f"and value should be one of this: {bound}"
    raise ValueError(f"'{name}' is a boolean {bound}.")


def check_tuple_int(name: str, values: tuple, bounds=None):
    if type(values) in [tuple, list] and len(values) > 1:
        value_flag = [type(item) == int for item in values]
        if np.all(value_flag):
            if bounds is not None and len(bounds) == len(values):
                value_flag = [is_in_bound(item, bound) for item, bound in zip(values, bounds)]
                if np.all(value_flag):
                    return values
            else:
                return values
    bounds = "" if bounds is None else f"and values should be in range: {bounds}"
    raise ValueError(f"'{name}' are integer {bounds}.")


def check_tuple_float(name: str, values: tuple, bounds=None):
    if type(values) in [tuple, list] and len(values) > 1:
        value_flag = [type(item) in [int, float] for item in values]
        if np.all(value_flag):
            if bounds is not None and len(bounds) == len(values):
                value_flag = [is_in_bound(item, bound) for item, bound in zip(values, bounds)]
                if np.all(value_flag):
                    return values
            else:
                return values
    bounds = "" if bounds is None else f"and values should be in range: {bounds}"
    raise ValueError(f"'{name}' are float {bounds}.")



#!/usr/bin/env python
# Created by "Thieu" at 16:25, 09/05/2023 ----------%                                                                               
#       Email: nguyenthieu2102@gmail.com            %                                                    
#       Github: https://github.com/thieu1995        %                         
# --------------------------------------------------%

# Paper: IHAOAVOA: An improved hybrid aquila optimizer and African vultures optimization algorithm for global optimization problems

import numpy as np
from enoppy.engineer import Engineer


class TensionCompressionSpringProblem(Engineer):
    """
    x = [x1, x2, x3] = [d, D, N]
    """

    name = "Tension/compression spring design problem"

    def __init__(self, f_penalty=None):
        super().__init__()
        self._n_dims = 3
        self._n_objs = 1
        self._n_cons = 4
        self._bounds = [(0.05, 2.0), (0.25, 1.3), (2.0, 15.0)]
        self.check_penalty_func(f_penalty)

    def get_objs(self, x):
        f1 = (x[2] + 2) * x[1]*x[0]**2
        return np.array([f1, ])

    def get_cons(self, x):
        g1 = 1 - (x[1]**3 * x[2]) / (71785 * x[0]**4)
        g2 = (4*x[1]**2 - x[0]*x[1])/(12566 * (x[1]*x[0]**3 - x[0]**4)) + 1. / (5108 * x[0]**2)
        g3 = 1 - 140.45*x[0] / (x[1]**2 * x[2])
        g4 = (x[0] + x[1]) / 1.5 - 1
        return np.array([g1, g2, g3, g4])

    def evaluate(self, x):
        self.n_fe += 1
        self.check_solution(x)
        list_objs = self.get_objs(x)
        list_cons = self.get_cons(x)
        return self.f_penalty(list_objs, list_cons)


class WeldedBeamProblem(Engineer):
    """
    x = [x1, x2, x3, x4] = [h, l, t, b]
    """

    name = "Welded beam design problem"

    def __init__(self, f_penalty=None):
        super().__init__()
        self._n_dims = 4
        self._n_objs = 1
        self._n_cons = 7
        self._bounds = [(0.1, 2.0), (0.1, 10.), (0.1, 10.), (0.1, 2.0)]
        self.L = 14
        self.E = 30*10**6
        self.G = 12*10**6
        self.theta_max = 0.25
        self.tau_max = 13600
        self.xichma_max = 30000
        self.P = 6000
        self.check_penalty_func(f_penalty)

    def get_objs(self, x):
        f1 = 1.10471 * x[0]**2 * x[1] + 0.04811 * x[2] * x[3] * (14 + x[1])
        return np.array([f1, ])

    def get_cons(self, x):
        Pc = 4.013*self.E*np.sqrt(x[2]**2 * x[3]**6 / 36) / self.L**2 * (1 - x[2]*np.sqrt(self.E/(4*self.G)) / (2*self.L))
        theta_z = 6 * self.P * self.L**3 / (self.E * x[2]**2 * x[3])
        xichma_z = 6*self.P*self.L / (self.E * x[2]**2 * x[3])
        jj = 2*np.sqrt(2)*x[0]*x[1]*(x[1]**2 / 4 + ((x[0] + x[2])/2)**2)
        R = np.sqrt(x[1]**2/4 + (x[0]+x[2])**2 / 4)
        M = self.P * (self.L + x[1]/2)
        tau2 = M * R / jj
        tau1 = self.P / (np.sqrt(2) * x[0]*x[1])
        tau = np.sqrt(tau1**2 + 2*tau1*tau2*x[1]/(2*R) + tau2**2)

        g1 = tau - self.tau_max
        g2 = xichma_z - self.xichma_max
        g3 = theta_z - self.theta_max
        g4 = x[0] - x[3]
        g5 = self.P - Pc
        g6 = 0.125 - x[0]
        g7 = 1.10471 * x[0]**2 + 0.04811 * x[2]*x[3]*(14 + x[1]) - 5
        return np.array([g1, g2, g3, g4, g5, g6, g7])

    def evaluate(self, x):
        self.n_fe += 1
        self.check_solution(x)
        list_objs = self.get_objs(x)
        list_cons = self.get_cons(x)
        return self.f_penalty(list_objs, list_cons)


class CantileverBeamProblem(Engineer):
    """
    x = [x1, x2, x3, x4, x5]
    """

    name = "Cantilever beam design problem"

    def __init__(self, f_penalty=None):
        super().__init__()
        self._n_dims = 5
        self._n_objs = 1
        self._n_cons = 1
        self._bounds = [(0.01, 100.0), ] * 5
        self.check_penalty_func(f_penalty)

    def get_objs(self, x):
        f1 = 0.6224 * np.sum(x)
        return np.array([f1, ])

    def get_cons(self, x):
        g1 = 61 / x[0]**3 + 27/x[1]**3 + 19/x[2]**3 + 7/x[3]**3 + 1/x[4]**3
        return np.array([g1, ])

    def evaluate(self, x):
        self.n_fe += 1
        self.check_solution(x)
        list_objs = self.get_objs(x)
        list_cons = self.get_cons(x)
        return self.f_penalty(list_objs, list_cons)


class SpeedReducerProblem(Engineer):
    """
    x = [x1, x2, x3, x4, x5, x6, x7]

    Ref: https://www.hindawi.com/journals/mpe/2013/419043/
    """

    name = "Speed reducer design problem"

    def __init__(self, f_penalty=None):
        super().__init__()
        self._n_dims = 7
        self._n_objs = 1
        self._n_cons = 11
        self._bounds = [(2.6, 3.6), (0.7, 0.8), (17, 28), (7.3, 8.3), (7.8, 8.3), (2.9, 3.9), (5.0, 5.5)]
        self.check_penalty_func(f_penalty)

    def get_objs(self, x):
        f1 = 0.7584*x[0]*x[1]**2*(3.3333*x[2]**2 + 14.9334*x[2]-43.0934) - 1.508*x[0]*(x[5]**2+x[6]**2) + 7.4777*(x[5]**3 + x[6]**3) + 0.7854*(x[3]*x[5]**2 + x[4]*x[6]**2)
        return np.array([f1, ])

    def get_cons(self, x):
        g1 = 27/(x[0]*x[1]**2*x[2]) - 1
        g2 = 397.5 / (x[0]*x[1]**2*x[2]**2) - 1
        g3 = 1.93*x[3]**3/ (x[1] * x[2]*x[5]**4) - 1
        g4 = 1.93 * x[4]**3 / (x[1] * x[2] * x[6]**4) - 1
        g5 = np.sqrt((745*x[3]/(x[1]*x[2]))**2 + 16.9*10**6) / (110*x[5]**3) - 1
        g6 = np.sqrt((745*x[4]/(x[1]*x[2]))**2 - 157.5*10**6) / (85 * x[6]**3) - 1
        g7 = x[1]*x[2]/40 - 1
        g8 = 5*x[1]/x[0] - 1
        g9 = x[0]/(12 * x[1]) - 1
        g10 = (1.5*x[5] + 1.9) / x[3] - 1
        g11 = (1.1*x[6] + 1.9) / x[4] -1
        return np.array([g1, g2, g3, g4, g5, g6, g7, g8, g9, g10, g11])

    def evaluate(self, x):
        self.n_fe += 1
        self.check_solution(x)
        list_objs = self.get_objs(x)
        list_cons = self.get_cons(x)
        return self.f_penalty(list_objs, list_cons)


class RollingElementBearingProblem(Engineer):
    """
    x = [x1, x2, x3, x4, x5, x6, x7, x8, x9, x10] = [Dm, Db, Z, fi, f0, Kdmin, Kdmax, theta, e, C]
    """

    name = "Rolling element bearing design problem"

    def __init__(self, f_penalty=None):
        super().__init__()
        self._n_dims = 10
        self._n_objs = 1
        self._n_cons = 9
        self.D = 160
        self.d = 90
        self.Bw = 30
        self.ri = self.ro = 11.033
        self._bounds = [(125., 150.), (10.5, 31.5), (4, 50), (0.515, 0.6), (0.515, 0.6), (0.4, 0.5), (0.6, 0.7), (0.3, 0.4), (0.02, 0.1), (0.6, 0.85)]
        self.check_penalty_func(f_penalty)

    def amend_position(self, x, lb=None, ub=None):
        x[2] = int(x[2])
        return x

    def get_objs(self, x):
        gama = x[1] / x[0]
        t1 = 37.91*(1+(1.04*((1-gama)/(1+gama))**1.72*(x[3]/x[4] * (2*x[4] - 1)/(2*x[3] - 1))**0.41)**(10./3))**(-0.3)
        fc = t1 * (gama**0.3 *(1 - gama)**1.39 / ((1+gama)**(1./3))) * (2*x[3]/(2*x[3] - 1))**0.41
        if x[1] <= 25.4:
            f1 = fc * x[2]**(2./3) * x[1]**1.8
        else:
            f1 = 3.647 * fc * x[2]**(2./3) * x[1]**1.4
        return np.array([f1, ])

    def get_cons(self, x):
        T = self.D - self.d - 2*x[1]
        xx = ((self.D - self.d) / 2 - 3 * (T/4))**2 + (self.D/2 - T/4 - x[1])**2 - (self.d/2 + T/4)**2
        yy = 2*((self.D - self.d)/2 - 3*T/4)*(self.D/2 - T/4 - x[1])
        theta0 = 2*np.pi - 1. / np.cos(xx / yy)

        g1 = theta0 / (2. / np.sin(x[1] / x[0])) - x[2] + 1
        g2 = x[5]*(self.D - self.d) - 2*x[1]
        g3 = 2*x[1] - x[6]*(self.D - self.d)
        g4 = x[-1]*self.Bw - x[1]
        g5 = 0.5*(self.D + self.d) - x[0]
        g6 = x[0] - (0.5 + x[-2])*(self.D + self.d)
        g7 = x[-3]*x[1] - 0.5*(self.D - x[0] - x[1])
        g8 = 0.515 - x[3]
        g9 = 0.515 - x[4]
        return np.array([g1, g2, g3, g4, g5, g6, g7, g8, g9])

    def evaluate(self, x):
        self.n_fe += 1
        self.check_solution(x)
        list_objs = self.get_objs(x)
        list_cons = self.get_cons(x)
        return self.f_penalty(list_objs, list_cons)


TCSP = TensionCompressionSpringProblem
WBP = WeldedBeamProblem
CBP = CantileverBeamProblem
SRP = SpeedReducerProblem
REBP = RollingElementBearingProblem
